module DaphneUtil
  module Tools

    require "pstore"
    require "cgi"
    require "json"
    require "rest-client"

    class Store

      attr_reader :loaded, :env

      def self.file_location
        "#{File.expand_path('~')}/.daphne"
      end

      def self.default_version
        "1"
      end

      def initialize
        @loaded = PStore.new(self.class.file_location, true)
        save
      end

      def env=(x_env)
        x_env = x_env.to_sym
        loaded.transaction do
          loaded[:envs].has_key?(x_env) ? (loaded[:env] = x_env) : x_env = nil
        end

        save
      end

      def add_env(name, x_url)
        new_env = {
            url: x_url,
            version: self.class.default_version,
            apps: {}
        }

        loaded.transaction { loaded[:envs][name.to_sym] = new_env }
        save
        self.env = name.to_sym
        loaded.transaction { loaded[:envs][name.to_sym] }
      end

      def rm_env(name)
        loaded.transaction do 

          loaded[:envs].delete(name.to_sym)
          loaded[:env] = nil if loaded[:envs][env].nil?
        end
        save
        loaded.transaction { loaded[:envs][env] }
      end

      def url
        loaded.transaction { loaded[:envs][env][:url] }
      end

      def url=(x_url)
        loaded.transaction do
          old_env = loaded[:envs][env]
          new_env = old_env.merge(url: x_url)

          loaded[:envs][env] = new_env
        end

        save
        loaded.transaction { loaded[:envs][env][:url] } 
      end

      def token
        loaded.transaction { loaded[:envs][env][:token] }
      end

      def token=(x_token)
        loaded.transaction do
          old_env = loaded[:envs][env]
          new_env = old_env.merge(token: x_token)

          loaded[:envs][env] = new_env
        end

        save
        loaded.transaction {loaded[:envs][env][:token]}
      end

      def version
        loaded.transaction { loaded[:envs][env][:version] }
      end

      def version=(x_version)
        loaded.transaction do
          old_env = loaded[:envs][env]
          new_env = old_env.merge(version: (x_version.scan(/\d+/).first || self.class.default_version))

          loaded[:envs][env] = new_env
        end

        save
        loaded.transaction {loaded[:envs][env][:version]}
      end

      def add_app(name, *seed)
        new_app = {name.to_sym => seed}
        loaded.transaction do
          old_apps = loaded[:envs][env][:apps]
          new_apps = old_apps.merge(new_app)
          loaded[:envs][env][:apps] = new_apps
        end

        save
        loaded.transaction {loaded[:envs][env][:apps]} 
      end

      def rm_app(name)
        loaded.transaction do
          old_env = loaded[:envs][env]
          new_env = old_env[:apps].delete(name.to_sym)
        end

        save

        loaded.transaction { loaded[:envs][name.to_sym] }
      end

      def apps
        loaded.transaction { loaded[:envs][env][:apps]}
      end

      def app_keys
        apps.keys
      end

      private

      def save
        loaded.transaction do
          loaded[:env] ||= :local
          loaded[:envs] ||= {}
          loaded[:envs][:local] ||= {url: "localhost:3001", version: "1", apps: {} }
          @env = loaded[:env]
          loaded.commit
        end
      end

    end

    def store
      @store ||= Store.new
    end

    def env
  %{  Environment #{store.env} :
      - url : #{store.url}
      - version : #{store.version}
      - token : #{store.token || 'No token set'}
      Credentials : #{apps}}
    end

    def app(k, v = nil)
      k = k.to_s.to_sym
      v ||= store.apps[k]
      seed = [k] + v
        %{          
          #{k} : #{v.join(',')}
              - uid : #{DaphneUtil.generate_id(*seed)}
              - secret : #{DaphneUtil.generate_secret(*seed)}
              - api_key : #{DaphneUtil.generate_api_key(*seed)}
}
    end

    def apps
      return 'None' if store.apps.empty?
      store.apps.collect {|k,v| app(k,v) }.join("\n")
    end

    def token
      store.token
    end

    def usage
      %{Environments
    create env <name> <url>
    delete env <name>
    get env
    env
Application Credentials
    create creds <name> [arg1, [arg2, [arg3] ...]]
Token
    token [string]
    verify <string>
    create token <creds name>
    create token <app uid> <app secret>
    create token <creds name> <email> <password>
    create token <app uid> <app secret> <email> <password>
Url
    url [string]
Application
    create app <attr1>=<value1> <attr2>=<value2> <attr3>=<value3>
    update app <uid or composite id> <attr1>=<value1> <attr2>=<value2> <attr3>=<value3>
    delete app <uid or composite id> 
    get app <uid or composite id>
Resource Owner
    create resource_owmer <attr1>=<value1> <attr2>=<value2> <attr3>=<value3>
    update resource_owmer <uid or composite id or email> <attr1>=<value1> <attr2>=<value2> <attr3>=<value3>
    delete resource_owmer <uid or composite id or email> 
    get resource_owmer <uid or composite id or email>
}
    end

    def api_domain
      "#{store.url}/api/v#{store.version}"
    end

    def oauth_domain
      "#{store.url}/oauth2/v#{store.version}"
    end

    def to_payload(args)
      r = CGI.parse(args.join('&'))
      Hash[*r.to_a.flatten]
    end

    def output(str)
      puts JSON.pretty_generate(JSON.parse(str))
    end

    def resource_mapping(key)
      {
        resource_owner: "resource_owners",
        application: "applications"
      }[key.to_s.to_sym]
    end

    def create_api_resource(resource_type, payload)
      raise IncorrectUsage unless resource_type && payload
      RestClient.post("#{api_domain}/#{resource_mapping(resource_type)}", {resource_type => payload}, authorization: "Bearer #{token}")
    end

    def update_api_resource(resource_type, identifier, payload)
      raise IncorrectUsage unless payload && identifier
      RestClient.put("#{api_domain}/#{resource_mapping(resource_type)}/#{identifier}", {resource_type => payload}, authorization: "Bearer #{token}")
    end

    def delete_api_resource(resource_type, identifier)
      raise IncorrectUsage unless identifier
      RestClient.delete("#{api_domain}/#{resource_mapping(resource_type)}/#{identifier}", authorization: "Bearer #{token}")
    end

    def retrieve_api_resource(resource_type, identifier)
      raise IncorrectUsage unless identifier
      RestClient.get("#{api_domain}/#{resource_mapping(resource_type)}/#{identifier}", authorization: "Bearer #{token}")
    end

    def create_token(*args)
      raise IncorrectUsage if args.empty?
      params = {grant_type: "client_credentials"}
      case args.length
      when 1
        if store.app_keys.include?(args.first.to_sym)
          seed = [args.first] + store.apps[args.first.to_sym]
          params[:client_id] = DaphneUtil.generate_id(*seed)
          params[:client_secret] = DaphneUtil.generate_secret(*seed)
        else
          raise IncorrectUsage
        end
      when 2
        params[:client_id] = args[0]
        params[:client_secret] = args[1]
      when 3
        raise IncorrectUsage
      else
        params[:grant_type] = "password"
        params[:client_id] = args[0]
        params[:client_secret] = args[1]
        params[:email] = args[2]
        params[:password] = args[3]
      end

      RestClient.post("#{oauth_domain}/token", params)
    end

    def verify_token(val)
      raise IncorrectUsage if val.to_s.empty?
      RestClient.post("#{oauth_domain}/verify?access_token=#{val}")
    end

    def create_env(name, url)
      raise IncorrectUsage unless url && name
      store.add_env(name, url)
    end

    def delete_env(name)
      raise IncorrectUsage unless name
      store.rm_env(name)
    end

    def create_app(name, *args)
      raise IncorrectUsage unless name
      store.add_app(name, *args)
    end

    def delete_app(name)
      raise IncorrectUsage unless name
      store.add_app(name)
    end

    def execute(*argv)
      raise IncorrectUsage if argv.empty?

      action, resource, args = argv[0], argv[1].to_s.to_sym, argv[2..-1].to_a

      case action
      when "create" then create(resource, *args)
      when "update" then update(resource, *args)
      when "delete" then delete(resource, *args)
      when "get"    then retrieve(resource, *args)
      when "verify" then verify_token(args[1])
      when "env" then retrieve(:env)
      when "help" then puts usage
      when "token"
        argv[1] ? update(:token, argv[1]) : retrieve(:token)
      when "url"
        argv[1] ? update(:url, argv[1]) : retrieve(:url)
      else raise IncorrectUsage
      end
    rescue IncorrectUsage
      puts usage
    rescue RestClient::Exception => e
      puts "Error :"
      e.is_a?(RestClient::InternalServerError) ? (raise e) : output(e.response)
    end

    def create(resource, *args)
      case resource
        when :resource_owner, :application, :app
          rtn = create_api_resource(resource, to_payload(args))
          output(rtn)
        when :token
          rtn = create_token(*args)
          output(rtn)
        when :env
          rtn = create_env(args[0], args[1])
          puts env
        when :creds
          create_app(args[0], *args[1..-1])
          puts env
        else raise IncorrectUsage
      end
    end

    def update(resource, *args)
      case resource
        when :resource_owner, :application, :app
          rtn = update_api_resource(resource, args[0], to_payload(args[1..-1]))
          output(rtn)
        when :token
          store.token = args.first
          puts env
        when :url
          store.url = args.first
          puts env
        else raise IncorrectUsage
      end
    end

    def delete(resource, *args)
      case resource
        when :resource_owner, :application, :app
          rtn = delete_api_resource(resource, args.first)
          output(rtn)
        when :env
          store.rm_env(args.first)
          puts env
        when :creds
          store.rm_app(args.first)
          puts env
        else raise IncorrectUsage
      end
    end

    def retrieve(resource, *args)
      case resource
        when :resource_owner, :application
          rtn = retrieve_api_resource(resource, args.first)
          output(rtn) 
        when :token
          puts(store.token ? store.token : "Token not configured for this environment.")
        when :env
          puts env
        else raise IncorrectUsage
      end
    end

  end
end